import os
import logging
import sys
import traceback
from typing import List
import click
import signal
from functools import partial

from logging.handlers import RotatingFileHandler


def terminate_application(application, *args):
    application.running = False


@click.group()
def cli():
    pass


@cli.command('visualize_log')
@click.option('--input-files', '-i', multiple=True, required=True, help='Input log files',
              type=click.Path(exists=True, dir_okay=False, file_okay=True))
@click.option('--plot_cpu', '-c', is_flag=True, required=False, type=bool,
              help='Plot CPU utilization statistics')
@click.option('--plot_memory', '-m', is_flag=True, required=False, type=bool,
              help='Plot memory utilization statistics')
@click.option('--plot_network', '-n', is_flag=True, required=False, type=bool,
              help='Plot network I/O')
@click.option('--plot_disk_io', '-d', is_flag=True, required=False, type=bool,
              help='Plot process\' disk I/O ')
@click.option('--plot_frame_drop', '-fd', is_flag=True, required=False, type=bool,
              help='Plot node\'s frame drop statistics')
@click.option('--plot_frame_rate', '-fr', is_flag=True, required=False, type=bool,
              help='Plot frame rate at node.')
@click.option('--plot_frame_io_time', '-fio', is_flag=True, required=False, type=bool,
              help='Plot frame I/O time at a node.')
@click.option('--plot_cumulative_node_latency', '-cl', is_flag=True, required=False, type=bool,
              help='Plot the time a frame took from its creation to the output of the specific '
                   'plugins.')
@click.option('--plot_node_latency', '-l', is_flag=True, required=False, type=bool, help='Plot the frames\' '
                                                                           'processing time at '
                                                                           'specific nodes')
def visualize_log(input_files: List[str],
                  plot_cpu: bool = False,
                  plot_memory: bool = False,
                  plot_network: bool = False,
                  plot_disk_io: bool = False,
                  plot_frame_drop: bool = False,
                  plot_frame_rate: bool = False,
                  plot_frame_io_time: bool = False,
                  plot_cumulative_node_latency: bool = True,
                  plot_node_latency: bool = True):
    from inds.core.profiler.csv_output import CsvOutput
    CsvOutput.visualize(input_files, plot_cpu=plot_cpu, plot_memory=plot_memory,
                        plot_network=plot_network, plot_disk_io=plot_disk_io,
                        plot_frame_drop=plot_frame_drop, plot_frame_rate=plot_frame_rate,
                        plot_frame_io_time=plot_frame_io_time,
                        plot_cumulative_node_latency=plot_cumulative_node_latency,
                        plot_node_latency=plot_node_latency)


@cli.command('run')
@click.option('--debug', is_flag=True, help='Set log level to DEBUG.')
def run(debug) -> None:
    print(''' __       ___          __   __   __   ___  ___ 
|  \  /\   |   /\     /__` |__) |__) |__  |__  
|__/ /--\  |  /--\    .__/ |    |  \ |___ |___

Inference DS
    ''')

    log_format = '%(asctime)s %(name)-32s %(levelname)-6s %(message)s'
    if debug:
        logging.basicConfig(level=logging.DEBUG, format=log_format)
    else:
        logging.basicConfig(level=logging.INFO, format=log_format)
        logging.getLogger('aiohttp.access').setLevel(logging.WARNING)

    application = None

    def new_except_hook(error_type, error_value, error_traceback) -> None:
        """
        Exception hook which logs uncaught exceptions to the logger output.

        @param error_type:  Type of the exception that is thrown but uncaught
        @param error_value: Instance of the exception that is thrown but uncaught, which contains
                            a specification on what actually happened.
        @param error_traceback:  Traceback, which we use to print debug information when available.
        """

        old_except_hook(error_type, error_value, error_traceback)

        format_traceback = '\n'.join([str(k) for k in traceback.walk_tb(error_traceback)])
        logging.getLogger().fatal(f'An unexpected error occurred:\n   Type: {error_type}\n  '
                                  f'Value: {error_value}\n   Traceback: \n{format_traceback}')

    old_except_hook = sys.excepthook
    sys.excepthook = new_except_hook

    try:

        from inds.core.utils import log_dir
        log_file = os.path.join(log_dir(), 'inds-core.log')
        rotating_file_handler = RotatingFileHandler(log_file, maxBytes=25 * 1024 * 1024,
                                                    backupCount=10)
        log_formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
        rotating_file_handler.setFormatter(log_formatter)
        logging.getLogger().addHandler(rotating_file_handler)

        from inds.core.application import Application
        application = Application()
        signal.signal(signal.SIGINT, partial(terminate_application, application))
        application.start()

    except KeyboardInterrupt:
        logging.info('Keyboard interrupt')
    except Exception as e:
        logging.error(f'An exception occurred: {e}')
        if debug:
            raise e

    if application is not None:
        logging.info('Stopping application...')
        application.stop()
    logging.info('Shutdown successful.')


if __name__ == '__main__':
    cli()
