import re
import inspect
from enum import Enum
from typing import Callable, List, Type, Dict, get_type_hints, Optional, Mapping, \
    Any, Tuple, Union, TypeVar, Set

from inds.core.exceptions import ConfigError


def camel_to_snake(name):
    """
    Transform snake case
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


def subclasses(super_class: Type):
    """
    Iteratively gather all descendents (allowing for multiple levels of inheritance) of the
    provided #super_class.
    """
    list_subclasses: List[Type] = super_class.__subclasses__()
    checked_types: Set[Type] = set()
    i: int = 0
    while i < len(list_subclasses):
        curr = list_subclasses[i]
        if curr not in checked_types:
            checked_types.add(curr)
            list_subclasses.extend(curr.__subclasses__())
        i += 1

    return set(list_subclasses)


def get_class(super_class: Type, sub_class_name: str, ignore_list=None) -> List:
    """
    Return list of sub classes of #super_class that match the #sub_class_name.
    If there is no naming ambiguity (*) , the returned list is either empty if the sub class is not 
    found, or contains exactly one entry.

    (*) Naming ambiguities can arise if the developers discard the naming convention, and provide 
    two sub classes of the same super_class, one of which following a camel case and the other 
    one a snake case naming convention.
    """
    if ignore_list is None:
        ignore_list = list()

    return [a for a in subclasses(super_class) if
            a not in ignore_list and camel_to_snake(a.__name__) == sub_class_name]


def is_user_defined_class(cls):
    return hasattr(cls, '__class__') and ('__dict__' in dir(cls) or hasattr(cls, '__slots__')
                                          or issubclass(cls, Enum))


def try_decode_type(value: Any, type_hint):
    is_container: bool = hasattr(type_hint, '__args__')
    decode_types: Tuple = type_hint.__args__ if is_container else (type_hint,)
    exceptions: List[Exception] = list()

    if type(value) == type_hint or type_hint == TypeVar or type_hint == Any:
        return True, value, []

    elif len(decode_types) == 1 and type_hint == List[decode_types]:
        if type(value) is list:
            ret: list = list()
            for val in value:
                success, sub_val, ex = try_decode_type(val, decode_types[0])
                if success:
                    ret.append(sub_val)
                else:
                    return False, ret, ex
            return True, ret, []

    elif len(decode_types) == 2 and type_hint == Dict[decode_types]:
        if type(value) is dict:
            ret: dict = dict()
            for k, v in value.items():
                success_key, sub_key, ex1 = try_decode_type(k, decode_types[0])
                success_val, sub_val, ex2 = try_decode_type(v, decode_types[1])
                if success_key and success_val:
                    ret[sub_key] = sub_val
                else:
                    return False, ret, ex1 + ex2
            return True, ret, []

    elif type_hint == Tuple[decode_types] and len(decode_types) > 1:
        if type(value) is list or type(value) is tuple and len(decode_types) == len(value):
            ret: list = list()
            for i, decode_type in enumerate(decode_types):
                success, sub_val, ex = try_decode_type(value[i], decode_type)
                if success:
                    ret.append(sub_val)
                else:
                    return False, ret, ex
            return True, tuple(ret), []

    elif type_hint == Union[decode_types] and len(decode_types) > 1:
        for decode_type in decode_types:
            success, val, exceptions = try_decode_type(value, decode_type)
            if success:
                return True, val, []
        return False, value, exceptions

    elif is_user_defined_class(type_hint):

        # If the user-defined type has a build method, try the build method can be used to
        # instantiate the instance of the class from the configuration
        if callable(getattr(type_hint, 'build', None)) and type(value) is dict:

            try:
                return True, type_hint.build(value), []
            except Exception as e:
                exceptions.append(e)

            if len(value) == 1 and type(list(value.keys())[0]) is str \
                    and type(list(value.values())[0]) is dict:
                try:
                    return True, type_hint.build(list(value.keys())[0], list(value.values())[0]), []
                except Exception as e:
                    exceptions.append(e)
        try:
            return True, type_hint(value), []
        except Exception as e:
            exceptions.append(e)

        if type(value) is str:
            subclass = get_class(type_hint, value)
            if len(subclass) == 1:
                try:
                    return True, subclass[0](), []
                except Exception as e:
                    exceptions.append(e)

        if type(value) is dict and len(value) == 1:
            subclass = get_class(type_hint, list(value.keys())[0])
            if len(subclass) == 1:
                return try_decode_type(list(value.values())[0], subclass[0])

    return False, value, exceptions


def parse_config_options(dictionary: Dict, function_ptr: Callable, variable_names: List):
    default_params: Mapping[str, inspect.Parameter] = inspect.signature(function_ptr).parameters
    k = function_ptr.__annotations__
    for k in function_ptr.__annotations__.keys():
        if function_ptr.__annotations__[k] == 'unicode':
            function_ptr.__annotations__[k] = 'str'

    type_hints: Dict[str, Type] = get_type_hints(function_ptr) # Can throw a type error

    return_values: Dict[str, Any] = dict()

    for variable_name in variable_names:

        # Fetch the default parameter if any
        default_parameter = default_params.get(variable_name)
        if default_parameter is not None and hasattr(default_parameter, 'default'):
            default_parameter = default_parameter.default

        # Fetch the user-defined value. In case the user did not specify the value in the
        # configuration dictionary, fall back to the default.
        value = dictionary.get(variable_name, default_parameter)

        # Check that the types match the required type.
        if variable_name in type_hints:
            type_hint = type_hints.get(variable_name)

            success, value, ex = try_decode_type(value, type_hint)
            if not success:
                ts: Tuple = type_hint.__args__ if hasattr(type_hint, '__args__') else (type_hint,)
                type_names = list(map(lambda a: a.__name__ if hasattr(a, '__name__') else a, ts))
                single_type_text = ('one of ' if len(ts) >= 1 else '')

                raise ConfigError(f'"{variable_name}" is defined as {value} and thus of type '
                                  f'{type(value)}. The type has to be {single_type_text} '
                                  f'{type_names}. {ex}')

        return_values[variable_name] = value

    return return_values


def instantiate_from_config(config: Optional[dict], class_handle: Type, *args, **kwargs):
    # Retrieve all parameters that are available to the class' constructor and erase those
    # that are specified manually.
    all_parameters = list(inspect.signature(class_handle.__init__, ).parameters)[1 + len(args):]
    for name in kwargs.keys():
        all_parameters.remove(name)

    cop: dict = dict()
    if config is not None:
        cop = parse_config_options(config, class_handle.__init__, list(all_parameters))

    return class_handle(*args, **cop, **kwargs)


class ConfigFieldTypes(str, Enum):
    GROUP = 'group'
    GROUP_SELECTION = 'group_selection'
    GROUP_SWITCHABLE = 'group_switchable'
    BOOL_INPUT = 'bool_input'
    DOUBLE_INPUT = 'double_input'
    INTEGER_INPUT = 'integer_input'
    STRING_INPUT = 'string_input'
    JSON_TEXT_INPUT = 'json_text_input'
    SELECTION_DROPDOWN = 'selection_dropdown'
    LIST_INPUT = 'list_input'

    def __json__(self):
        return self.value


MAX_QUEUE_SIZE_CONFIG_FIELD = {
    'name': 'max_queue_size',
    'type': ConfigFieldTypes.INTEGER_INPUT,
    'display_name': 'max. queue size',
    'default': 1,
    'description': 'Maximum size of the input queue of the node.'
}


def parse_max_queue_size(config):
    max_queue_size = config.get('max_queue_size', 1)
    if max_queue_size is None or type(max_queue_size) != int or max_queue_size <= 0:
        raise ConfigError('max_queue_size must be an integer that is larger than zero')
    return max_queue_size


RESIZE_CONFIG_FIELD = {
    'name': 'resize',
    'type': ConfigFieldTypes.GROUP_SWITCHABLE,
    'display_name': 'resize',
    'description': 'Resize images.',
    'fields': [
        {
            'name': 'width',
            'type': ConfigFieldTypes.INTEGER_INPUT,
            'display_name': 'width'
        },
        {
            'name': 'height',
            'type': ConfigFieldTypes.INTEGER_INPUT,
            'display_name': 'height'
        }
    ],
}


def parse_resize_field(config):
    resize = config.get('resize')
    if resize is not None:
        if type(resize) != dict or resize.get('width') is None or resize.get('height') is None:
            raise ConfigError('resize must contain the entries "width" and "height"')
        if resize.get('width') < 1.0 or resize.get('height') < 1.0:
            raise ConfigError('resize width and height must be larger than 1.0')
    return resize


TRIGGER_CONFIG_FIELD = \
    {
        'name': 'trigger',
        'type': ConfigFieldTypes.GROUP_SELECTION,
        'display_name': 'trigger',
        'description': 'Trigger for processing images of the node.',
        'default': 'default_trigger',
        'groups': [
            {
                'name': 'default_trigger',
                'type': ConfigFieldTypes.GROUP,
                'display_name': 'default',
                'description': 'Triggers for each image.',
                'fields': []
            },
            {
                'name': 'never_trigger',
                'type': ConfigFieldTypes.GROUP,
                'display_name': 'never',
                'description': 'Triggers never.',
                'fields': []
            },
            {
                'name': 'frame_skip_trigger',
                'type': ConfigFieldTypes.GROUP,
                'display_name': 'frame skip',
                'description': 'Triggers and skips the specified number of frames before the next '
                               'frame is triggered.',
                'fields': [{
                    'name': 'skip_frames',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'display_name': 'skip frames',
                    'default': 0,
                    'description': 'Number of frames to skip'
                }]
            },
            {
                'name': 'difference_trigger',
                'type': ConfigFieldTypes.GROUP,
                'display_name': 'difference trigger',
                'description': 'Computed the difference between the triggered image and the next '
                               'image. If the difference exceeds the specified threshold, the '
                               'frame is triggered.',
                'fields': [{
                    'name': 'threshold',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'display_name': 'threshold',
                    'default': 10,
                    'description': 'Difference threshold'
                }]
            }
        ],
    }

BATCH_AGGREGATION_ATTRIBUTES_FIELD = \
    {
        'name': 'attributes',
        'display_name': 'Attributes',
        'type': ConfigFieldTypes.LIST_INPUT,
        'description': 'Add the path of all attributes that are to be aggregated and stored '
                       'with the batch output.',
        'fields': [
            {
                'name': 'attribute',
                'display_name': 'Complete Attribute Path (optional)',
                'type': ConfigFieldTypes.LIST_INPUT,
                'fields': [
                    {
                        'name': 'attribute',
                        'display_name': 'Attribute Name (optional)',
                        'type': ConfigFieldTypes.STRING_INPUT
                    }
                ]
            },
        ]
    }
