import logging
from abc import ABC, abstractmethod
from typing import Dict, Optional, List

from inds.core.consumer import Consumer
from inds.core.decorators import action
from inds.core.producer import Producer
from inds.core.config import camel_to_snake


class Node(ABC):
    def __init__(self, name: str) -> None:
        self.logger = logging.getLogger(__name__)

        self.name = name
        self.consumers: Dict[str, Consumer] = dict()
        self.producers: Dict[str, Producer] = dict()
        self.running: bool = False

        self.actions: Dict[str, callable] = dict()
        for name in dir(self.__class__):
            func = getattr(self.__class__, name)
            if isinstance(func, action):
                self.actions[name] = func

    def add_consumer(self, consumer: Consumer):
        self.consumers[consumer.name] = consumer

    def add_producer(self, producer: Producer):
        self.producers[producer.name] = producer

    def start(self):
        self.running = True
        self.run()

    def perform_action(self, action_name: str, args: List, kwargs: Dict):
        try:
            func = getattr(self, action_name)
        except AttributeError:
            return

        if func is not None:
            func(self, *args, **kwargs)

    @staticmethod
    def get_config_entries() -> Dict:
        """Return the entries of the configuration of this kind of nodes."""
        return {}

    @staticmethod
    @abstractmethod
    def get_config_categories() -> List[str]:
        """Return the categories of the configuration of this kind of nodes."""

    @staticmethod
    def get_config_key() -> Optional[str]:
        """
        Node implementations can implement this function by providing the configuration key
        that identifies their class in the configuration file.
        If this function function returns None, the default convention is used, which
        utilizes the class' name in camel case.
        """
        return None

    @classmethod
    def config_key(cls) -> str:
        key = cls.get_config_key()
        return camel_to_snake(cls.__name__) if key is None else key

    @staticmethod
    @abstractmethod
    def build(name: str, config: Dict) -> Optional['Node']:
        """Create an instance of the node from a configuration."""

    @abstractmethod
    def run(self):
        """Run the processing loop."""

    def stop(self):
        """Stop the node."""
        self.running = False

    def disconnect(self):
        """Remove all connections to other nodes (producers and consumers)."""
        for consumer in self.consumers.values():
            consumer.disconnect_all_producers()

        for producer in self.producers.values():
            producer.disconnect_all_consumers()
