import numpy as np

from inds.core.decorators import onjit


def iou(boxes1, n_boxes1, boxes2):
    """
    :param boxes1
    :param n_boxes1:
    :param boxes2
    :return: IoU
    """

    batch_size = boxes1.shape[0]
    max_n_gt = np.max(n_boxes1)
    n_anchors = boxes2.shape[0]
    ious = np.zeros((batch_size, max_n_gt, n_anchors))

    for i in range(batch_size):
        n_gt = n_boxes1[i]

        for j in range(n_gt):
            x1 = np.expand_dims(boxes1[i, j, 0], 0)
            y1 = np.expand_dims(boxes1[i, j, 1], 0)
            l1 = np.expand_dims(boxes1[i, j, 2], 0)  # length in x direction
            w1 = np.expand_dims(boxes1[i, j, 3], 0)  # width in y direction

            x2 = np.expand_dims(boxes2[:, 0], 0)
            y2 = np.expand_dims(boxes2[:, 1], 0)
            l2 = np.expand_dims(boxes2[:, 2], 0)
            w2 = np.expand_dims(boxes2[:, 3], 0)

            dx = np.abs(x1 - x2.T)
            dy = np.abs(y1 - y2.T)

            l_intersect = np.minimum(np.maximum(0.0, 0.5 * l1 + 0.5 * l2.T - dx), np.minimum(l1, l2.T))
            w_intersect = np.minimum(np.maximum(0.0, 0.5 * w1 + 0.5 * w2.T - dy), np.minimum(w1, w2.T))

            area1 = np.multiply(l1, w1)
            area2 = np.multiply(l2, w2)
            area_intersect = np.multiply(l_intersect, w_intersect)

            area_union = area1 + area2.T - area_intersect
            intersection_over_union = np.divide(area_intersect, area_union,
                                                out=np.zeros_like(area_intersect),
                                                where=area_union != 0).T
            ious[i, j, :] = intersection_over_union

    return ious


def iou_single(boxes1, boxes2, matching_threshold):
    """
    Compute the IoU of the two given boxes.
    :param boxes1
    :param boxes2
    :param orientation: Compute IoU of oriented boxes
    :return: IoU
    """

    if boxes1.ndim == 1:
        boxes1_n = np.expand_dims(boxes1, 0)
    else:
        boxes1_n = boxes1

    x1 = np.expand_dims(boxes1_n[:, 0], 0)
    y1 = np.expand_dims(boxes1_n[:, 1], 0)
    w1 = np.expand_dims(boxes1_n[:, 2], 0)
    h1 = np.expand_dims(boxes1_n[:, 3], 0)

    if boxes2.ndim == 1:
        boxes2_n = np.expand_dims(boxes2, 0)
    else:
        boxes2_n = boxes2

    x2 = np.expand_dims(boxes2_n[:, 0], 0)
    y2 = np.expand_dims(boxes2_n[:, 1], 0)
    w2 = np.expand_dims(boxes2_n[:, 2], 0)
    h2 = np.expand_dims(boxes2_n[:, 3], 0)

    dx = np.abs(x1 - x2.T)
    dy = np.abs(y1 - y2.T)

    w_intersect = np.minimum(np.maximum(0.0, 0.5 * w1 + 0.5 * w2.T - dx), np.minimum(w1, w2.T))
    h_intersect = np.minimum(np.maximum(0.0, 0.5 * h1 + 0.5 * h2.T - dy), np.minimum(h1, h2.T))

    area1 = np.multiply(w1, h1)
    area2 = np.multiply(w2, h2)
    area_intersect = np.multiply(w_intersect, h_intersect)

    area_union = area1 + area2.T - area_intersect

    intersection_over_union = np.divide(area_intersect, area_union, out=np.zeros_like(area_intersect), where=area_union != 0).T

    return np.array(intersection_over_union) > matching_threshold


@onjit
def iou_single_njit(box1, box2, matching_threshold):
    """
    Compute the IoU of the two given boxes.
    :param box1
    :param box2
    :param matching_threshold:
    :return: IoU
    """

    x1 = box1[0]
    y1 = box1[1]
    w1 = box1[2]
    h1 = box1[3]

    x2 = box2[0]
    y2 = box2[1]
    w2 = box2[2]
    h2 = box2[3]

    dx = np.abs(x1 - x2)
    dy = np.abs(y1 - y2)

    w_intersect = np.minimum(np.maximum(0.0, 0.5 * w1 + 0.5 * w2 - dx), np.minimum(w1, w2))
    h_intersect = np.minimum(np.maximum(0.0, 0.5 * h1 + 0.5 * h2 - dy), np.minimum(h1, h2))

    area1 = np.multiply(w1, h1)
    area2 = np.multiply(w2, h2)
    area_intersect = np.multiply(w_intersect, h_intersect)

    area_union = area1 + area2 - area_intersect

    if area_union != 0:
        return area_intersect / area_union > matching_threshold
    else:
        return False


@onjit
def euclidean_distance_njit(box1, box2, matching_threshold):
    """
    Compute the IoU of the two given boxes.
    :param box1
    :param box2
    :param matching_threshold:
    :return: euclidean distance between both boxes
    """

    dx = box1[0] - box2[0]
    dy = box1[1] - box2[1]

    return np.sqrt(dx*dx + dy*dy) < matching_threshold
