from inds.core.decorators import onjit

import numpy as np
import inds.inference.matching as matching


@onjit
def multi_class_non_maximum_suppression(confidences, boxes, classes, n_classes, max_selections=200, iou_threshold=0.5):
    """Apply non maximum suppression to the objects of each class separately

    :param confidences: Matrix (numpy) of shape [N, 1] containing the confidence for each box
    :param boxes: Matrix (numpy) of shape [N, 4]
    :param classes: Matrix (numpy) of shape [N, C] containing the class for each box
    :param n_classes: Number of classes
    :param max_selections: Maximum number of selected boxes per class
    :param iou_threshold: Threshold when to merge boxes
    :return: Selected box indices
    """

    all_selected_indices = []

    for i in range(1, n_classes + 1):
        class_i_indices = np.where(classes == i)[0]
        class_boxes = boxes[class_i_indices]
        class_confidences = confidences[class_i_indices]

        if class_i_indices.shape[0] > 0:
            selected_indices = non_maximum_suppression(class_confidences, class_boxes, max_selections, iou_threshold)
            all_selected_indices.extend(class_i_indices[selected_indices])

    return all_selected_indices


@onjit
def non_maximum_suppression(confidences, boxes, max_selections=200, matching_threshold=0.6, matching_fn=matching.iou_single_njit):
    """Apply non maximum suppression to aggregate highly overlapping boxes.
    :param confidences: Matrix (numpy) of shape [N, 1] containing the confidence for each box
    :param boxes: Matrix (numpy) of shape [N, 4]
    :param max_selections: Maximum number of selected boxes
    :param matching_threshold: Threshold when to merge boxes
    :param matching_fn: Function to use for comparing two boxes. Use IoU as default.
    :return: Selected box indices
    """

    # sort (descending)
    sorted_indices = np.argsort(confidences)[::-1]
    n_select = min(boxes.shape[0], max_selections)
    selected_indices = np.zeros(n_select, dtype=np.int32)

    # boxes_valid = matching.iou_single(boxes, boxes) > iou_threshold

    n_selected = 0
    for i in sorted_indices:
        select = True

        # iterate backwards through selected boxes (overlapping boxes are likely to have similar scores)
        # see implementation of NMS in tensorflow/core/kernels/non_max_suppression_op.cc
        if n_selected > 0:
            # select = not np.any(matching.iou_single(boxes[i], boxes[selected_indices]) > iou_threshold)
            # select = not (np.any(boxes_valid[i, selected_indices[:n_selected]]))

            for j in selected_indices[n_selected-1::-1]:
                if matching_fn(boxes[i], boxes[j], matching_threshold):
                    select = False
                    break

        if select:
            selected_indices[n_selected] = i
            n_selected += 1
            if n_selected >= n_select:
                break

    return selected_indices[:n_selected]
